# {{ description or name.replace('_', ' ').capitalize() }}
{
  "{{ config.AI_SYNTAX_FUNCTION_NAME_FIELD }}": "{{ name }}"{% if args %}{{ "," }}{% endif %}
{%- for k,v in args.items() %}
  "{{ k }}":
    {%- if v.type %} <{{ v.type }}>{% endif -%}
    {%- if v.default != 'NOT_SET' %} (default = {{ v.default }}){%endif-%}
    {%- if not loop.last -%},{%- endif -%}
    {%- if v.comment %} {{ v.comment }}{% endif -%}
{%- endfor -%}
{{ "\n}" }}